@@ -27,7 +27,6 @@ from cognitivefactory.interactive_clustering.clustering.abstract import (  # To
 from cognitivefactory.interactive_clustering.constraints.abstract import (  # To manage constraints.
     AbstractConstraintsManager,
 )
-from cognitivefactory.interactive_clustering.utils import checking  # To check parameters.
 
 
 # ==============================================================================
@@ -42,7 +41,6 @@ class KMeansConstrainedClustering(AbstractConstrainedClustering):
         - KMeans Clustering: `MacQueen, J. (1967). Some methods for classification and analysis of multivariate observations. Proceedings of the fifth Berkeley symposium on mathematical statistics and probability 1(14), 281–297.`
         - Constrained _'COP'_ KMeans Clustering: `Wagstaff, K., C. Cardie, S. Rogers, et S. Schroedl (2001). Constrained K-means Clustering with Background Knowledge. International Conference on Machine Learning`
 
-
     Examples:
         ```python
         # Import.
@@ -70,8 +68,8 @@ class KMeansConstrainedClustering(AbstractConstrainedClustering):
             "8": csr_matrix([0.00, 0.00, 0.00, 1.00]),
         }
 
-        # Define constraints manager (set it to None for no constraints).
-        constraints_manager = BinaryConstraintsManager(list_of_data_IDs=["0", "1", "2", "3", "4", "5", "6", "7", "8"])
+        # Define constraints manager.
+        constraints_manager = BinaryConstraintsManager(list_of_data_IDs=list(vectors.keys()))
         constraints_manager.add_constraint(data_ID1="0", data_ID2="1", constraint_type="MUST_LINK")
         constraints_manager.add_constraint(data_ID1="0", data_ID2="7", constraint_type="MUST_LINK")
         constraints_manager.add_constraint(data_ID1="0", data_ID2="8", constraint_type="MUST_LINK")
@@ -81,9 +79,9 @@ class KMeansConstrainedClustering(AbstractConstrainedClustering):
 
         # Run clustering.
         dict_of_predicted_clusters = clustering_model(
-            vectors=vectors
-            nb_clusters=3
-            constraints_manager=constraints_manager
+            constraints_manager=constraints_manager,
+            vectors=vectors,
+            nb_clusters=3,
         )
 
         # Print results.
@@ -146,9 +144,9 @@ class KMeansConstrainedClustering(AbstractConstrainedClustering):
     # ==============================================================================
     def cluster(
         self,
+        constraints_manager: AbstractConstraintsManager,
         vectors: Dict[str, Union[ndarray, csr_matrix]],
         nb_clusters: int,
-        constraints_manager: Optional[AbstractConstraintsManager] = None,
         verbose: bool = False,
         **kargs,
     ) -> Dict[str, int]:
@@ -156,9 +154,9 @@ class KMeansConstrainedClustering(AbstractConstrainedClustering):
         The main method used to cluster data with the KMeans model.
 
         Args:
-            vectors (Dict[str,Union[ndarray,csr_matrix]]): The representation of data vectors. The keys of the dictionary represents the data IDs. This keys have to refer to the list of data IDs managed by the `constraints_manager` (if `constraints_manager` is set). The value of the dictionary represent the vector of each data. Vectors can be dense (`numpy.ndarray`) or sparse (`scipy.sparse.csr_matrix`).
-            nb_clusters (int): The number of clusters to compute. #TODO Set defaults to None with elbow method or other method ?
-            constraints_manager (Optional[AbstractConstraintsManager], optional): A constraints manager over data IDs that will force clustering to respect some conditions during computation. The list of data IDs managed by `constraints_manager` has to refer to `vectors` keys. If `None`, no constraint are used during the clustering. Defaults to `None`.
+            constraints_manager (AbstractConstraintsManager): A constraints manager over data IDs that will force clustering to respect some conditions during computation.
+            vectors (Dict[str,Union[ndarray,csr_matrix]]): The representation of data vectors. The keys of the dictionary represents the data IDs. This keys have to refer to the list of data IDs managed by the `constraints_manager`. The value of the dictionary represent the vector of each data. Vectors can be dense (`numpy.ndarray`) or sparse (`scipy.sparse.csr_matrix`).
+            nb_clusters (int): The number of clusters to compute.  #TODO Set defaults to None with elbow method or other method ?
             verbose (bool, optional): Enable verbose output. Defaults to `False`.
             **kargs (dict): Other parameters that can be used in the clustering.
 
@@ -173,26 +171,16 @@ class KMeansConstrainedClustering(AbstractConstrainedClustering):
         ### GET PARAMETERS
         ###
 
-        # Get `list_of_data_IDs`.
+        # Store `self.constraints_manager` and `self.list_of_data_IDs`.
+        if not isinstance(constraints_manager, AbstractConstraintsManager):
+            raise ValueError("The `constraints_manager` parameter has to be a `AbstractConstraintsManager` type.")
+        self.constraints_manager: AbstractConstraintsManager = constraints_manager
+        self.list_of_data_IDs: List[str] = self.constraints_manager.get_list_of_managed_data_IDs()
+
+        # Store `self.vectors`.
         if not isinstance(vectors, dict):
             raise ValueError("The `vectors` parameter has to be a `dict` type.")
-        self.list_of_data_IDs: List[str] = sorted(vectors.keys())
-        if not isinstance(self.list_of_data_IDs, list) or not all(
-            isinstance(element, str) for element in self.list_of_data_IDs
-        ):
-            raise ValueError("The `list_of_data_IDs` variable has to be a `list` type.")
-
-        # Check `constraints_manager`.
-        self.constraints_manager: AbstractConstraintsManager = checking.check_constraints_manager(
-            list_of_data_IDs=self.list_of_data_IDs,
-            constraints_manager=constraints_manager,
-        )
-
-        # Check `vectors`.
-        self.vectors: Dict[str, Union[ndarray, csr_matrix]] = checking.check_vectors(
-            list_of_data_IDs=self.list_of_data_IDs,
-            vectors=vectors,
-        )
+        self.vectors: Dict[str, Union[ndarray, csr_matrix]] = vectors
 
         # Store `self.nb_clusters`.
         if nb_clusters < 2:
@@ -293,8 +281,16 @@ class KMeansConstrainedClustering(AbstractConstrainedClustering):
                     cluster_ID for cluster_ID in self.centroids.keys() if cluster_ID not in not_compatible_cluster_IDs
                 ]
 
+                # If there is no possible cluster...
+                if possible_cluster_IDs == []:  # noqa: WPS520
+
+                    # Assign the data ID to a new cluster
+                    new_clusters[data_ID_to_assign] = max(
+                        max([cluster_ID for _, cluster_ID in new_clusters.items()]) + 1, self.nb_clusters
+                    )
+
                 # If there is possible clusters...
-                if possible_cluster_IDs:
+                else:
 
                     # Compute distance between data ID and all possible centroids.
                     distances_to_cluster_ID: Dict[float, int] = {
@@ -310,14 +306,6 @@ class KMeansConstrainedClustering(AbstractConstrainedClustering):
                     min_distance: float = min(distances_to_cluster_ID)
                     new_clusters[data_ID_to_assign] = distances_to_cluster_ID[min_distance]
 
-                # If there is not possible cluster...
-                else:
-
-                    # Assign the data ID to a new cluster
-                    new_clusters[data_ID_to_assign] = max(
-                        max([cluster_ID for _, cluster_ID in new_clusters.items()]) + 1, self.nb_clusters
-                    )
-
                 # Assign all data ID that are linked by a `"MUST_LINK"` constraint to the new cluster.
                 for data_ID in self.list_of_data_IDs:
                     if (
@@ -331,9 +319,6 @@ class KMeansConstrainedClustering(AbstractConstrainedClustering):
                             list_of_data_IDs_to_assign.remove(data_ID)
                         new_clusters[data_ID] = new_clusters[data_ID_to_assign]
 
-            # Rename cluster IDs by order.
-            new_clusters = rename_clusters_by_order(clusters=new_clusters)
-
             ###
             ### COMPUTE NEW CENTROIDS
             ###
@@ -373,10 +358,10 @@ class KMeansConstrainedClustering(AbstractConstrainedClustering):
             ###
 
             # Affect new clusters.
-            self.clusters = new_clusters.copy()
+            self.clusters = new_clusters
 
             # Affect new centroids.
-            self.centroids = new_centroids.copy()
+            self.centroids = new_centroids
 
             # Verbose.
             if verbose and (current_iteration % 5 == 0):  # pragma: no cover
@@ -388,6 +373,10 @@ class KMeansConstrainedClustering(AbstractConstrainedClustering):
             # Verbose - Print progression status.
             print("    CLUSTERING_ITERATION=" + str(current_iteration), ",", "converged=" + str(converged))
 
+        # Rename cluster IDs by order.
+        self.clusters = rename_clusters_by_order(clusters=self.clusters)
+        self.centroids = self.compute_centroids(clusters=new_clusters)
+
         return self.clusters
 
     # ==============================================================================
